@@ -80,7 +80,7 @@ int daemonize_fd = -1;
 
 /* --monitor: Should a supervisory process monitor the daemon and restart it if
  * it dies due to an error signal? */
-static bool monitor;
+bool monitor;
 
 /* --user: Only root can use this option. Switch to new uid:gid after
  * initially running as root.  */
@@ -435,8 +435,8 @@ monitor_daemon(pid_t daemon_pid)
 /* If daemonization is configured, then starts daemonization, by forking and
  * returning in the child process.  The parent process hangs around until the
  * child lets it know either that it completed startup successfully (by calling
- * daemon_complete()) or that it failed to start up (by exiting with a nonzero
- * exit code). */
+ * daemonize_complete()) or that it failed to start up (by exiting with a
+ * nonzero exit code). */
 void
 daemonize_start(bool access_datapath)
 {
