@@ -77,7 +77,7 @@ lldpd_get_hardware(struct lldpd *cfg, char *name, int index,
 }
 
 struct lldpd_hardware *
-lldpd_alloc_hardware(struct lldpd *cfg, char *name, int index)
+lldpd_alloc_hardware(struct lldpd *cfg, const char *name, int index)
 {
     struct lldpd_hardware *hw;
 
@@ -244,6 +244,7 @@ lldpd_decode(struct lldpd *cfg, char *frame, int s,
 
     if (s < sizeof(struct eth_header) + 4) {
         /* Too short, just discard it */
+        hw->h_rx_discarded_cnt++;
         return;
     }
 
@@ -284,6 +285,7 @@ lldpd_decode(struct lldpd *cfg, char *frame, int s,
                 VLOG_DBG("function for %s protocol did not "
                          "decode this frame",
                          cfg->g_protocols[i].name);
+                hw->h_rx_discarded_cnt++;
                 return;
             }
             chassis->c_protocol = port->p_protocol = cfg->g_protocols[i].mode;
@@ -408,8 +410,6 @@ lldpd_decode(struct lldpd *cfg, char *frame, int s,
     if (!oport)  {
         hw->h_insert_cnt++;
     }
-
-    return;
 }
 
 static void
