@@ -470,6 +470,7 @@ do_ca_cert_bootstrap(struct stream *stream)
 static char *
 get_peer_common_name(const struct ssl_stream *sslv)
 {
+    char *peer_name = NULL;
     X509 *peer_cert = SSL_get_peer_certificate(sslv->ssl);
     if (!peer_cert) {
         return NULL;
@@ -478,18 +479,18 @@ get_peer_common_name(const struct ssl_stream *sslv)
     int cn_index = X509_NAME_get_index_by_NID(X509_get_subject_name(peer_cert),
                                               NID_commonName, -1);
     if (cn_index < 0) {
-        return NULL;
+        goto error;
     }
 
     X509_NAME_ENTRY *cn_entry = X509_NAME_get_entry(
         X509_get_subject_name(peer_cert), cn_index);
     if (!cn_entry) {
-        return NULL;
+        goto error;
     }
 
     ASN1_STRING *cn_data = X509_NAME_ENTRY_get_data(cn_entry);
     if (!cn_data) {
-        return NULL;
+        goto error;
     }
 
     const char *cn;
@@ -499,7 +500,11 @@ get_peer_common_name(const struct ssl_stream *sslv)
 #else
     cn = (const char *)ASN1_STRING_get0_data(cn_data);
  #endif
-    return xstrdup(cn);
+    peer_name = xstrdup(cn);
+
+error:
+    X509_free(peer_cert);
+    return peer_name;
 }
 
 static int
@@ -757,18 +762,22 @@ ssl_send(struct stream *stream, const void *buffer, size_t n)
     if (sslv->txbuf) {
         return -EAGAIN;
     } else {
+        struct ofpbuf buf;
         int error;
 
-        sslv->txbuf = ofpbuf_clone_data(buffer, n);
+        ofpbuf_use_const(&buf, buffer, n);
+        sslv->txbuf = &buf;
         error = ssl_do_tx(stream);
         switch (error) {
         case 0:
-            ssl_clear_txbuf(sslv);
+            sslv->txbuf = NULL;
             return n;
         case EAGAIN:
+            /* Copy remaining data. */
+            sslv->txbuf = ofpbuf_clone_data(buf.data, buf.size);
             return n;
         default:
-            ssl_clear_txbuf(sslv);
+            sslv->txbuf = NULL;
             return -error;
         }
     }
@@ -1082,7 +1091,6 @@ tmp_dh_callback(SSL *ssl OVS_UNUSED, int is_export OVS_UNUSED, int keylength)
     };
 
     static struct dh dh_table[] = {
-        {1024, NULL, get_dh1024},
         {2048, NULL, get_dh2048},
         {4096, NULL, get_dh4096},
     };
@@ -1090,7 +1098,7 @@ tmp_dh_callback(SSL *ssl OVS_UNUSED, int is_export OVS_UNUSED, int keylength)
     struct dh *dh;
 
     for (dh = dh_table; dh < &dh_table[ARRAY_SIZE(dh_table)]; dh++) {
-        if (dh->keylength == keylength) {
+        if (dh->keylength >= keylength) {
             if (!dh->dh) {
                 dh->dh = dh->constructor();
                 if (!dh->dh) {
