@@ -31,7 +31,6 @@
 #include "simap.h"
 #include "hash.h"
 #include "table.h"
-#include "hash.h"
 #include "timeval.h"
 #include "transaction.h"
 #include "jsonrpc-server.h"
@@ -406,6 +405,10 @@ ovsdb_monitor_columns_sort(struct ovsdb_monitor *dbmon)
     SHASH_FOR_EACH (node, &dbmon->tables) {
         struct ovsdb_monitor_table *mt = node->data;
 
+        if (mt->n_columns == 0) {
+            continue;
+        }
+
         qsort(mt->columns, mt->n_columns, sizeof *mt->columns,
               compare_ovsdb_monitor_column);
         for (i = 0; i < mt->n_columns; i++) {
@@ -1232,6 +1235,15 @@ ovsdb_monitor_get_update(
                                             condition,
                                             ovsdb_monitor_compose_row_update2);
                 if (!condition || !condition->conditional) {
+                    if (json) {
+                        struct json *json_serialized;
+
+                        /* Pre-serializing the object to avoid doing this
+                         * for every client. */
+                        json_serialized = json_serialized_object_create(json);
+                        json_destroy(json);
+                        json = json_serialized;
+                    }
                     ovsdb_monitor_json_cache_insert(dbmon, version, mcs,
                                                     json);
                 }
