@@ -26,7 +26,6 @@
 #include "command-line.h"
 #include "daemon.h"
 #include "dirs.h"
-#include "dns-resolve.h"
 #include "openvswitch/dynamic-string.h"
 #include "fatal-signal.h"
 #include "file.h"
@@ -330,7 +329,6 @@ main(int argc, char *argv[])
     service_start(&argc, &argv);
     fatal_ignore_sigpipe();
     process_init();
-    dns_resolve_init(true);
 
     bool active = false;
     parse_options(argc, argv, &db_filenames, &remotes, &unixctl_path,
@@ -513,7 +511,6 @@ main(int argc, char *argv[])
                       run_command, process_status_msg(status));
         }
     }
-    dns_resolve_destroy();
     perf_counters_destroy();
     service_stop();
     return 0;
@@ -907,8 +904,9 @@ query_db_string(const struct shash *all_dbs, const char *name,
 
             datum = &row->fields[column->index];
             for (i = 0; i < datum->n; i++) {
-                if (datum->keys[i].string[0]) {
-                    return datum->keys[i].string;
+                const char *key = json_string(datum->keys[i].s);
+                if (key[0]) {
+                    return key;
                 }
             }
         }
@@ -1021,7 +1019,7 @@ query_db_remotes(const char *name, const struct shash *all_dbs,
 
             datum = &row->fields[column->index];
             for (i = 0; i < datum->n; i++) {
-                add_remote(remotes, datum->keys[i].string);
+                add_remote(remotes, json_string(datum->keys[i].s));
             }
         }
     } else if (column->type.key.type == OVSDB_TYPE_UUID
